/************************************************************************/
/* CW_Trainer1.c - Morse code reviewer program.				*/
/* For info, see:  https://sourceforge.net/projects/cw-trainer/		*/
/*									*/
/* Compile on linux with:                  				*/
/*  cc cw_trainer1.c -lm -o cw_trainer1.exe				*/
/*									*/
/* Run as:								*/
/*  cw_trainer1.exe							*/
/* 									*/
/* For help and to see options, invoke with -h or -help.		*/
/*									*/
/* Copyright (C) 2001 - Carl Kindman					*/
/*                                                                      */
/* GNU Public License - GPL:						*/
/* This program is free software; you can redistribute it and/or	*/
/* modify it under the terms of the GNU General Public License as	*/
/* published by the Free Software Foundation; either version 2 of the	*/
/* License, or (at your option) any later version.			*/
/*									*/
/* This program is distributed in the hope that it will be useful,	*/
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU	*/
/* General Public License for more details.				*/
/*									*/
/* You should have received a copy of the GNU General Public License	*/
/* along with this program; if not, write to the Free Software		*/
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA		*/
/* 02111-1307 USA							*/
/*									*/
/* Carl Kindman - 6-8-01						*/
/************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <math.h>
#include <time.h>


float WPMrate=5.0;		/* Rate in Words Per Minute (WPM), 			*/
				/*  Where, avgerage word is expected to have 5-letters,	*/
				/*  and average letter has 2 da's and 2 di's.		*/
				/* The di and interclick durations are equal and become	*/
				/* the atomic time unit.  The da equals 3 di's.		*/
				/* So the average letter lasts 12 units + 6 inter-	*/
				/* letter gap, or 18 unit times per letter.  Which 	*/
				/* gives 5*18=90 units per word + 12 inter-word gap, 	*/
				/* gives 102 time units per word.  So base time unit is	*/
				/* 60 seconds / 102 = 0.59 seconds for 1-WPM.		*/
				/* 5-WPM implies 0.59/5 = 0.12 second time unit.	*/
float TimeUnit=0.59;

float farnsworth_compression = 1.75;
				/* Farnsworth compression, shortens the symbol-clicking */
				/* and lengthens the space between letters for easier	*/
				/* learning.  The default timings (above) contain a 	*/
				/* small amount of Farnsworth compression, by virtue of */
				/* the "6-inter-letter" gap, which would otherwise be	*/
				/* 3.  You can further increase the gap/symbol ratio by	*/
				/* choosing a higher Farnsworth_Compression value.	*/
				/* For example, a value of 2, doubles the ratio.	*/
				/* (Default ratio=6/12=0.5 .)				*/


char Letters[41][10] =		/* Values:  ' '=silent, '.'="di", '-'="da".	*/
				/* Index 0 is 'space'.			*/
				/* Indices 1-26 are letters a-z.	*/
				/* Indices 27-36 are numbers 0-9.	*/
  /* Pre-Initialize code alphabet. */
 { "      ",	/* space */
   ".-",	/* A 	1 */
   "-...",	/* B 	2 */
   "-.-.",	/* C 	3 */
   "-..",	/* D 	4 */
   ".",		/* E 	5 */
   "..-.",	/* F 	6 */
   "--.",	/* G 	7 */
   "....",	/* H 	8 */
   "..",	/* I 	9 */
   ".---",	/* J 	10 */
   "-.-",	/* K 	11 */
   ".-..",	/* L 	12 */
   "--",	/* M 	13 */
   "-.",	/* N 	14 */
   "---",	/* O 	15 */
   ".--.",	/* P 	16 */
   "--.-",	/* Q 	17 */
   ".-.",	/* R 	18 */
   "...",	/* S 	19 */
   "-",		/* T 	10 */
   "..-",	/* U 	21 */
   "...-",	/* V 	22 */
   ".--",	/* W 	23 */
   "-..-",	/* X 	24 */
   "-.--",	/* Y 	25 */
   "--..",	/* Z 	26 */
   "-----",	/* 0 	27 */
   ".----",	/* 1 	28 */
   "..---",	/* 2 	29 */
   "...--",   	/* 3 	30 */
   "....-",	/* 4 	31 */
   ".....",	/* 5 	32 */
   "-....",	/* 6 	33 */
   "--...",	/* 7 	34 */
   "---..",	/* 8 	35 */
   "----.",	/* 9 	36 */
   ".-.-.-",	/* . 	37 */
   "--..--",	/* , 	38 */
   "..--..",	/* ? 	39 */
   " "    	/* short-space */
 };


#define maxrecord (2 * 6 * 10 * 48000)	/* Allow up to 3-second tone at 48kHz. */

short keydown[maxrecord],	/* Provide max of 3 seconds of "da". */
      keyup[  maxrecord];	/* Provide max of 3 seconds of space. */

int tones_not_intialized=1;


void send_symbol( char *symb )
{
 short x;
 int j, n, m;
 int Hzrate = 8000;
 int format = AFMT_S16_LE;
 int chan=2; /* can't go lower than stereo on some systems */
 int outfile;
 float freq=1000.0;  /* 1 KHz. */
 float dphi;

 outfile = open("/dev/audio", O_WRONLY);
 if (outfile==-1) printf("ERROR: Cannot open audio file output.\n");
 if (ioctl(outfile,SNDCTL_DSP_SETFMT, &format)==-1) { perror("format");exit(1); }
 if (ioctl(outfile,SNDCTL_DSP_SPEED, &Hzrate)==-1) { perror("speed error");exit(1); }
 if (ioctl(outfile,SNDCTL_DSP_CHANNELS, &chan)==-1) { perror("channels error");exit(1); }

 if (tones_not_intialized)
  {
   /* Initialize signal segments. */
   dphi = 2.0 * 3.14159 * freq / (float)Hzrate;
   for (j=0; j<maxrecord/2; j++)
    {
     x = 30000 * sin( (float)j * dphi );
     keydown[j*2] = x;
     keydown[j*2+1] = x;
     keyup[j*2] = 0;
     keyup[j*2+1] = 0;
    }
   tones_not_intialized = 0;
  }

 n = 2.0 * TimeUnit * (float)Hzrate / farnsworth_compression;
 if (3 * n >= maxrecord) {printf("Warning:  Max duration exceeded (n=%d).\n", n);  n = maxrecord/3 - 1; }
 m = 2.0 * TimeUnit * (float)Hzrate * farnsworth_compression;
 if (6 * m >= maxrecord) {printf("Warning:  Max duration exceeded (m=%d vs. %d).\n", m, maxrecord/6);  m = maxrecord/6 - 1; }

 j = 0;
 while (symb[j] != '\0')
  {
   switch (symb[j])
    {
     case ' ':	/* silent. */
		write(outfile, keyup, 2*m);
	break;
     case '.':	/* Di. */
		write(outfile, keydown, 2*n);
	break;
     case '-':	/* Da. */
		write(outfile, keydown, 3 * 2*n);
	break;
     default: printf("Error: Unexpected symbol %x\n", symb[j] );
    }
   write(outfile, keyup, 2*n);	/* Place inter-klick gap. */
   j = j + 1; 
  }
 write(outfile, keyup, 6 * 2*m);	/* Place inter-letter gap. */
 close(outfile);
}


void Play_Random( int CW_Mode )
{
 char word[100];
 int j, k;

 if (CW_Mode==1)  /* Wait for correct keyboard response, before generating each next symbol. */
  {
   int correct=1;

   send_symbol( Letters[40] );
   while (1)
    {
     if (correct)
      j = 1 + 39.0 * (float)random()/(float)RAND_MAX;	/* Pick a new letter. */
     correct = 1;
     send_symbol( Letters[j] );
     scanf("%s",word);
     word[0] = toupper(word[0]);
     k = word[0];
     if (j<27) { if (k!=j+64) { correct = 0; printf("%cWrong: %c\n", 7, j+64); } } else
     if (j<37) { if (k-48!=j-27) { correct = 0; printf("%cWrong: %c\n", 7, j+48-27); } } else
     if (j==37) { if (k!='.') { correct = 0; printf("%cWrong: .\n", 7); } } else
     if (j==38) { if (k!=',') { correct = 0; printf("%cWrong: ,\n", 7); } } else
     { if (k!='?') { correct = 0; printf("%cWrong: ?\n", 7); } } 
     /* send_symbol( Letters[40] ); */
     usleep( 100000 );
    }
  }

 else

 if (CW_Mode==3)  /* Continuous groups, letters and numbers, without waiting for response. */
  {
   int correct=1;

   send_symbol( Letters[40] );
   while (1)
    {
     j = 1 + 39.0 * (float)random()/(float)RAND_MAX;	/* Pick a new letter or number. */
     if (j<27)  printf("%c\n", j+64+32);  else
     if (j<37)  printf("%c\n", j+48-27);  else
     if (j==37)  printf(".\n");  else
     if (j==38)  printf(",\n");  else  printf("?\n");
     send_symbol( Letters[j] );
     if ((float)random()/(float)RAND_MAX > 0.8) { printf("\n");  send_symbol( Letters[40] ); }
     else usleep( 100000 );
    }
  }

 else

 if (CW_Mode==2)  /* Continuous groups, letters only, without waiting for response. */
  {
   int correct=1;
   char ch;

   send_symbol( Letters[40] );
   while (1)
    {
     j = 1 + 26.0 * (float)random()/(float)RAND_MAX;	/* Pick a new letter. */
     ch = j+64+32;
     if (j<27)  printf("%c\n", j+64+32);  else
     if (j<37)  printf("%c\n", j+48-27);  else
     if (j==37)  printf(".\n");  else
     if (j==38)  printf(",\n");  else  printf("?\n");
     send_symbol( Letters[j] );
     if ((float)random()/(float)RAND_MAX > 0.8) { printf("\n");  send_symbol( Letters[40] ); }
     else usleep( 100000 );
    }
  }
}




void Play_Text( char *text )
{ /*specific_letters*/
  int j=0;

  while (text[j]!='\0')
   {
    printf("	%c\n", text[j]);
    text[j] = toupper(text[j]);
    if ((text[j]>64) && (text[j]<91))
     send_symbol( Letters[ text[j] - 64 ] );
    else
    if ((text[j]>47) && (text[j]<58))
     send_symbol( Letters[ 26 + text[j] - 47 ] );
    else
    if (text[j] == ' ')
     send_symbol( Letters[0] );
    else
    if (text[j] == '.')
     send_symbol( Letters[37] );
    else
    if (text[j] == ',')
     send_symbol( Letters[38] );
    else
    if (text[j] == '?')
     send_symbol( Letters[39] );
    else printf("No code for '%c'\n", text[j] );
    j++;
   }
} /*specific_letters*/


void Play_File( char *file_name )
{ 
  char ch;
  FILE *infile;

  infile = fopen(file_name,"r");
  if (infile==0) { printf("Cannot open file '%s' for reading.\n", file_name);  exit(0); }
  ch = getc(infile);
  while (!feof(infile))
   {
    printf("	%c\n", ch);
    ch = toupper(ch);
    if ((ch>64) && (ch<91))
     send_symbol( Letters[ ch - 64 ] );
    else
    if ((ch>47) && (ch<58))
     send_symbol( Letters[ 26 + ch - 47 ] );
    else
    if (ch == ' ')
     send_symbol( Letters[0] );
    else
    if (ch == '.')
     send_symbol( Letters[37] );
    else
    if (ch == ',')
     send_symbol( Letters[38] );
    else
    if (ch == '?')
     send_symbol( Letters[39] );
    else printf("No code for '%c'\n", ch );
    ch = getc(infile);
   }
 printf("Done reading file '%s'\n", file_name);
} 



void safe_gets( char *word, int n )
{
 int k=0;

 do{
 word[k++] = getchar();
printf("comnparing %d to %d\n", word[k-1], '\n');
}
 while ((!feof(stdin)) && (k<n-1) && (word[k-1]!='\n'));
 word[k-1] = '\0';
}



/****************************************************************/
/* Main -							*/
/****************************************************************/
int main(int argc, char *argv[])
{
  int i, j, k, rand_generate_mode=0;
  char filename[1000]="", word[5000];

 printf("\nCW Trainer  1.0 -\n\n");
 srandom( time(0) );

  j = 1;
  while (j < argc)	/* Parse any command-line arguments. */
   {
    if (strcmp(argv[j],"-wpm")==0)
     {
	j++;
	if (sscanf(argv[j],"%f",&WPMrate)!=1) printf("Error: Bad float WPM Rate '%s'.\n", argv[j]);
     }
    else
    if (strcmp(argv[j],"-fwc")==0)
     {
	j++;
	if (sscanf(argv[j],"%f",&farnsworth_compression)!=1) printf("Error: Bad float Farnsworth Compression Ratio '%s'.\n", argv[j]);
     }
    else
    if (strstr(argv[j],"-rand2")==argv[j])
     {
	rand_generate_mode = 2;
     }
    else
    if (strstr(argv[j],"-rand3")==argv[j])
     {
	rand_generate_mode = 3;
     }
    else
    if (strstr(argv[j],"-rand")==argv[j])
     {
	rand_generate_mode = 1;
     }
    else
    if (strstr(argv[j],"-h")==argv[j])
     {
	printf("Command Options Summary:\n");
	printf("  -wpm xx     - Set Words-Per-Minute speed.  Default = 5.0 wpm.\n");
	printf("  -fwc xx     - Set Farnsworth Compression ratio. Default=1.75.\n");
	printf("  -rand       - Sends symbols randomly, and wait for correct\n");
	printf("                keyboard response before generating next symbol.\n");
	printf("                (Good for new beginners.  Go at your own pace.)\n");
	printf("  -rand2      - Sends continuous groups of letters-only, without\n");
	printf("                waiting for keyboard response. (Use paper&pencil.)\n");
	printf("                (Good for intermediate.)\n");
	printf("  -rand3      - Sends continuous groups of letters, numbers, and\n");
	printf("                punctuation.  No waiting for responses.\n");
	printf("                (Good for experts.)\n");
	printf("  'text-file' - Reads the text-file and sends the contents.\n");
	printf("                (Good for pre-prepared practice on realistic text.)\n");
	printf("  <nothing>   - Prompts for text interactively and sends it.\n");
	printf("                Where <nothing> is not -randx or text-file,\n"); 
	printf("                except -wpm and -fwc.\n");
	printf("                (Good for newest beginners learning sounds interactively.\n\n");
	exit(0);
    }
     else
     {
	strcpy(filename, argv[j]);	/* Read text from a file. */
     }
    j = j + 1;
   }

 printf("CW-Trainer	Version 1.0	(Invoke with -h for help.)\n");
 printf("Rate set to %g WPM. (Set with -wpm xx)\n", WPMrate);
 printf("Farnsworth Compression Ratio set to %g. (Set with -fwc xx)\n", farnsworth_compression);
 printf("\n");

 TimeUnit = TimeUnit / WPMrate;

 switch (rand_generate_mode)
  {
   case 0:  /* Use supplied text. (Not random.) */
	    if (strlen(filename) > 0)  Play_File( filename );
	    else
	     {	/* Accept text from keyboard. */
		printf("Enter text to send as CW-code.  (End by ^D or ^C.)\nText? ");
		safe_gets( word, 5000 );
		while (!feof(stdin))
		 {
		  Play_Text(word);
		  printf("Text? ");
		  safe_gets(word, 5000 );
		 }
	     }
	break;

   case 1:	printf("Generating random text (%d).\n", rand_generate_mode);
		printf("Enter character and press <enter> after each symbol before continuing.\n");
		Play_Random( rand_generate_mode );

   default:  	printf("Generating random text (%d).\n", rand_generate_mode);
		Play_Random( rand_generate_mode );
  }

}
